module io4d.iotype;

public import std.c.time;
public import std.stdint;
public import std.c.stdarg;
public import std.socket;
public import std.bitmanip;

extern(C):

struct timezone {
    int tz_minuteswest; /* of Greenwich */
    int tz_dsttime;     /* type of dst correction to apply */
}

// basekit
struct List {}
struct UArray {}
alias UArray PtrUArray;
struct BStream {}
struct FILE {}
struct RandomGen {}
struct PointerHash {}
struct PointerHashRecord {}
struct CHashRecord {}
struct CHash {}
struct Stack {}
struct MainArgs {
    int argc;
    const(char*)* argv;
}
struct DynLib {}
struct Date {}
struct Duration {}
struct Collector {}
struct CollectorMarker {
    CollectorMarker *prev;
    CollectorMarker *next;
    uint color; // : 2
    uint hash1;
    uint hash2;
    IoObjectData* object;
}
alias int CTYPE;
enum : CTYPE {
    CTYPE_uint8_t,
    CTYPE_uint16_t,
    CTYPE_uint32_t,
    CTYPE_uint64_t,

    CTYPE_int8_t,
    CTYPE_int16_t,
    CTYPE_int32_t,
    CTYPE_int64_t,

    CTYPE_float32_t,
    CTYPE_float64_t,

    CTYPE_uintptr_t
}
alias size_t ssize_t;
alias int CENCODING;
enum : CENCODING {
    CENCODING_ASCII,
    CENCODING_UTF8,
    CENCODING_UCS2,
    CENCODING_UCS4,
    CENCODING_NUMBER
}


// coro
struct Coro{}
alias void* function(void*) CoroStartCallback;



// garbagecollector
alias int function(CollectorMarker*) CollectorCheckFunc;
alias int function(CollectorMarker*) CollectorMarkFunc;
alias void function(CollectorMarker*) CollectorWillFreeFunc;
alias void function(CollectorMarker*) CollectorFreeFunc;


// vm
struct IoState {
    RandomGen* randomGen;
    PointerHash* primitives;
    CHash* symbols;

    // coroutines

    IoObject* objectProto;
    IoObject* mainCoroutine;    // the object that represents the main "thread"
    IoObject* currentCoroutine; // the object whose coroutine is active
    Stack* currentIoStack;      // quick access to current coro's retain stack

    // quick access objects

    IoSymbol *activateSymbol;
    IoSymbol *callSymbol;
    IoSymbol *forwardSymbol;
    IoSymbol *noShufflingSymbol;
    IoSymbol *opShuffleSymbol;
    //IoSymbol *performSymbol;
    //IoSymbol *referenceIdSymbol;
    IoSymbol *semicolonSymbol;
    IoSymbol *selfSymbol;
    IoSymbol *setSlotSymbol;
    IoSymbol *setSlotWithTypeSymbol;
    IoSymbol *stackSizeSymbol;
    IoSymbol *typeSymbol;
    IoSymbol *updateSlotSymbol;

    IoObject *setSlotBlock;
    IoObject *localsUpdateSlotCFunc;
    IoObject *localsProto;

    IoMessage *asStringMessage;
    IoMessage *collectedLinkMessage;
    IoMessage *compareMessage;
    //IoMessage *doStringMessage;
    IoMessage *initMessage;
    IoMessage *mainMessage;
    IoMessage *nilMessage;
    IoMessage *opShuffleMessage;
    IoMessage *printMessage;
    IoMessage *referenceIdForObjectMessage;
    IoMessage *objectForReferenceIdMessage;
    IoMessage *runMessage;
    IoMessage *willFreeMessage;
    IoMessage *yieldMessage;
    IoMessage *didFinishMessage;

    List *cachedNumbers;

    // singletons

    IoObject *ioNil;
    IoObject *ioTrue;
    IoObject *ioFalse;

    // Flow control singletons

    IoObject *ioNormal;
    IoObject *ioBreak;
    IoObject *ioContinue;
    IoObject *ioReturn;
    IoObject *ioEol;

    // garbage collection

    Collector *collector;
    IoObject *lobby;
    IoObject *core;

    // recycling

    List *recycledObjects;
    size_t maxRecycledObjects;

    // startup environment

    MainArgs *mainArgs;

    // current execution state

    int stopStatus;
    void *returnValue;

    // embedding

    void *callbackContext;
    IoStateBindingsInitCallback bindingsInitCallback;
    IoStatePrintCallback        printCallback;
    IoStateExceptionCallback    exceptionCallback;
    IoStateExitCallback         exitCallback;
    IoStateActiveCoroCallback   activeCoroCallback;

    // debugger

    int debugOn;
    IoObject *debugger;
    IoMessage *vmWillSendMessage;

    // SandBox limits

    int messageCountLimit;
    int messageCount;
    double timeLimit;
    double endTime;

    // exiting

    int shouldExit;
    int exitResult;

    int receivedSignal;
    int showAllMessages;

    //CHash *profiler;
}



struct IoTag {
    void* state;
    char* name;

    // memory management

    IoTagCloneFunc cloneFunc;
    IoTagFreeFunc freeFunc;
    IoTagCleanupFunc tagCleanupFunc;
    IoTagMarkFunc markFunc;
    IoTagNotificationFunc notificationFunc;

    // actions

    //IoTagTouchFunc touchFunc; // if present, call before type check
    IoTagPerformFunc performFunc; // lookup and activate, return result
    IoTagActivateFunc activateFunc; // return the receiver or compute and return a value
    IoTagCompareFunc compareFunc;

    /*
      IoTagSetFunc setFunc;
      IoTagGetFunc getFunc;
      IoTagCursorFunc cursorFunc;
    */

    // persistence

    IoTagWriteToStreamFunc   writeToStreamFunc;
    IoTagAllocFromStreamFunc allocFromStreamFunc;
    IoTagReadFromStreamFunc  readFromStreamFunc;

    // Stack* recyclableInstances;
    // int maxRecyclableInstances;
    int referenceCount;
}


struct IoObjectData {
    uint markerCount;
    union data_ {
        union pointers_ {
            void* ptr;
            void* ptr2;
        }
        pointers_ pointers;
        double d;
        uint ui32;
    } // 8 bytes (9+8 = 17)
    data_ data;
    
    // 4*4 = 16 bytes (17+16 = 33)
    IoTag* tag;
    PHash* slots;
    List* listeners;
    IoObject** protos;

    mixin(bitfields!(uint, "hasDoneLookup", 1,
                     uint, "isActivatable", 1,
                     uint, "isDirty",       1,
                     uint, "ownsSlots",     1,
                     uint, "isSymbol",      1,
                     uint, "isLocals",      1,
                     uint, "isReferenced",  1,
                     uint, "sentWillFree",  1));
    /+
     unsigned int hasDoneLookup : 1;    // used to avoid slot lookup loops
     unsigned int isActivatable : 1;    // if true, upon activation, call activate slot
     unsigned int isDirty : 1;          // set to true when the object changes its storable state
     
     // optimizations
     
     unsigned int ownsSlots : 1;        // if true, io_free slots hash table when io_freeing object
     unsigned int isSymbol : 1;         // true if the object is a unqiue sequence - a symbol
     unsigned int isLocals : 1;         // true if the Object is a locals object
     unsigned int isReferenced : 1;     // 1 bit ref counter
     unsigned int sentWillFree : 1;     // 1 is sent willFree mes
     //unsigned int isPObject : 1;        // check before type tests?
     +/
}
// "typedef" for error message
typedef CollectorMarker IoObject;
alias IoObject IoCFunction;
alias IoObject IoMessage;
alias IoObject IoCoroutine;
alias IoObject IoNumber;
alias IoObject IoSeq;
alias IoObject IoSymbol;
alias IoObject IoDirectory;
alias IoObject IoList;
alias IoObject IoMap;
alias IoObject IoFile;
struct PHash {}
struct PHashRecord {}
struct IoMethodTable {}
struct vec2f { float x, y; }
struct vec3f { float x, y, z; }



// self, locals, message
alias IoObject* function(IoObject*, IoObject*, IoMessage*) IoUserFunction;
alias IoUserFunction IoMethodFunc; // (IoObject*, IoObject*, IoObject*)

alias IoUserFunction IoTagPerformFunc; // (void*, void*, void*)
alias IoObject* function(IoObject*, IoObject*, IoObject*, IoMessage*, void*) IoTagActivateFunc;
alias int function(IoObject*, IoObject*) IoTagCompareFunc;
alias void function(IoObject*, BStream*) IoTagWriteToStreamFunc;
alias void* function(IoObject*, BStream*) IoTagAllocFromStreamFunc;
alias void function(IoObject*, BStream*) IoTagReadFromStreamFunc;
alias IoObject* function(IoObject*) IoTagCloneFunc;
alias void function(IoObject*) IoTagFreeFunc;
alias void function(IoObject*) IoTagMarkFunc;
alias void function(IoObject*, void*) IoTagNotificationFunc;
alias void function(IoObject*) IoTagCleanupFunc;

alias intptr_t function(void*) CHashHashFunc;
alias int function(void*, void*) CHashEqualFunc;

alias void function(void*) ListDoCallback;
alias void function(void*, void*) ListDoWithCallback;
alias void* function(void*) ListCollectCallback;
alias int function(void*) ListSelectCallback;
alias int function(void*) ListDetectCallback;
alias int function(const(void)*, const(void)*) ListSortCallback;
alias int function(void*, const(void)*, const(void)*) ListSortRCallback;
alias int function(void*, const(void)*, const(void)*) PortableSortingCompareCallback;

alias void function(void*) StackDoCallback;
alias void function(void*, void*) StackDoOnCallback;

alias int function(const(void)*, const(void)*) UArraySortCallback;

alias IoObject* function(IoState*) IoStateProtoFunc;
alias void function(IoState*, const(UArray)*) IoStatePrintCallback;
alias void function(IoState*, int) IoStateExitCallback;
alias void function(IoState*, IoObject*) IoStateExceptionCallback;
alias void function(IoState*, void*) IoStateBindingsInitCallback;
alias void function(IoState*, int) IoStateActiveCoroCallback;



