module io4d.iowrap;

public import io4d.iovmapi;
import std.traits, std.stdio;
import std.string;
import std.conv;
import std.exception;
import std.c.string : strcmp;
import std.typecons;
import core.memory;
import std.typetuple;




/// Io Macros
alias Collector_makeGrayIfWhite_ Collector_shouldMark_;
///ditto
IoObject* IOTRUE(IoObject* self) {
    return IoObject_state(self).ioTrue;
}
///ditto
IoState* IOSTATE(IoObject* self) {
    return IoObject_state(self);
}
/// ditto
Collector* IOCOLLECTOR(IoState* state) {
    return state.collector;
}
/// ditto
bool ISTRUE(IoObject* self) {
    return self !is IoObject_state(self).ioNil
        && self !is IoObject_state(self).ioFalse;
}
/// ditto
IoObject* IOFALSE(IoObject* self) {
    return IoObject_state(self).ioFalse;
}
/// ditto
int ISFALSE(IoObject* self) {
    return self is IoObject_state(self).ioNil
        || self is IoObject_state(self).ioFalse;
}
/// ditto
IoObject* IOBOOL(IoObject* self, bool b) {
    return b ? IOTRUE(self) : IOFALSE(self);
}
/// ditto
IoTag* IoObject_tag(IoObject* self) {
    return IoObject_deref(self).tag;
}
/// ditto
void IoObject_tag_(IoObject* self, IoTag* t) {
    return (IoObject_deref(self).tag = t);
}
/// ditto
IoObjectData* IoObject_deref(IoObject* self) {
    return (cast(CollectorMarker*)self).object;
}
/// ditto
bool ISNUMBER(IoObject* self) {
    return strcmp(IoObject_name(self), "Number") == 0;
    //return (IoObject_hasCloneFunc_(self, cast(IoTagCloneFunc)&IoNumber_rawClone) != 0);
}
/// ditto
bool ISSEQ(IoObject* self) {
    return strcmp(IoObject_name(self), "Sequence") == 0;
    //return (IoObject_hasCloneFunc_(self, cast(IoTagCloneFunc)&IoSeq_rawClone) != 0);
}
/// ditto
bool ISBOOL(IoObject* self) {
    auto state = IoObject_state(self);
    return self is state.ioFalse
        || self is state.ioNil
        || self is state.ioTrue;
}
/// ditto
IoState* IoObject_state(IoObject* self) {
    return cast(IoState*)IoObject_tag(self).state;
}
/// ditto
void* IoObject_dataPointer(IoObject* self) {
    return IoObject_deref(self).data.pointers.ptr;
}
/// ditto
void IoObject_setDataPointer_(IoObject* self, void* p) {
    IoObject_deref(self).data.pointers.ptr = p;
}
/// ditto
void IoObject_isDirty_(IoObject* self, uint v) {
    IoObject_deref(self).isDirty = v;
}
/// ditto
uint IoObject_isDirty(IoObject* self) {
    return IoObject_deref(self).isDirty;
}
/// ditto
IoSymbol* IOSYMBOL(IoState* self, string s) {
    return IoState_symbolWithCString_length_(self, s.ptr, s.length);
}

/// ditto
void IOASSERT(IoObject* self, IoMessage* m, bool b, lazy string str) {
    if (!b) {
        IoState_error_(IoObject_state(self),
                       m,
                       "Io Assertion '%s'",
                       str.toStringz());
        assert(0);
    }
}

/*
template IoDefineEnum(string name, T...) {
    
}
*/

/// 列挙体CTYPEを型から導きだす
template CTypeOf(T) {
    static if (is(Unqual!T == ubyte) || is(Unqual!T == void))
        alias CTYPE_uint8_t CTypeOf;
    else static if (is(Unqual!T == byte))
        alias CTYPE_int8_t CTypeOf;
    else static if (is(Unqual!T == short))
        alias CTYPE_int16_t CTypeOf;
    else static if (is(Unqual!T == ushort))
        alias CTYPE_uint16_t CTypeOf;
    else static if (is(Unqual!T == int))
        alias CTYPE_int32_t CTypeOf;
    else static if (is(Unqual!T == uint))
        alias CTYPE_uint32_t CTypeOf;
    else static if (is(Unqual!T == long))
        alias CTYPE_int64_t CTypeOf;
    else static if (is(Unqual!T == ulong))
        alias CTYPE_uint64_t CTypeOf;
    else static if (is(Unqual!T == float))
        alias CTYPE_float32_t CTypeOf;
    else static if (is(Unqual!T == double))
        alias CTYPE_float64_t CTypeOf;
    else static if (is(Unqual!T == char))
        alias CTYPE_uint8_t CTypeOf;
    else static if (is(Unqual!T == wchar))
        alias CTYPE_uint16_t CTypeOf;
    else static if (is(Unqual!T == dchar))
        alias CTYPE_uint32_t CTypeOf;
    else static if (is(Unqual!T == bool))
        alias CTYPE_uint8_t CTypeOf;
    else
        static assert(0, "'" ~ T.stringof ~ "' is unimplemented");
}
/// 列挙体CENCODINGを型から導きだす
template CEncodingOf(T) {
    static if (is(Unqual!T == char))
        alias CENCODING_UTF8 CEncodingOf;
    else static if (is(Unqual!T == wchar))
        alias CENCODING_UCS2 CEncodingOf;
    else static if (is(Unqual!T == dchar))
        alias CENCODING_UCS4 CEncodingOf;
    else
        alias CENCODING_NUMBER CEncodingOf; // binary data
}


/**
   IoObjectをDの型に変換する

   Example:
       -----
       IoObject* piobj = IoState_doCString(state, "1 atan * 4");
       double pi = IoObjectTo!double(piobj);
       writeln(pi);
       -----
 */
T IoObjectTo(T)(IoObject* self) {
    static if (isNumeric!T) {
        if (ISNUMBER(self)) {
            static if (is(T == float))
                return IoNumber_asFloat(self);
            else static if (is(T == double))
                return IoNumber_asDouble(self);
            else
                return IoNumber_asInt(self);
        }
        else if (ISSEQ(self)) {
            return cast(T)IoSeq_asDouble(self);
        }
        else {
            IoState_error_(IoObject_state(self), null,
                           "Unsupported type '%s'", IoObject_name(self));
            assert(0);
        }
    }
    else static if (is(Unqual!T == bool)) {
        //if (ISBOOL(self)) {
        return ISTRUE(self);
        //}
        //else {
        //    IoState_error_(IoObject_state(self), null,
        //                   "IoObjectTo!bool() got not boolean value. '%s'", IoObject_name(self));
        //    assert(0);
        //}
    }
    else static if (is(Unqual!T == char*)) {
        if (ISSEQ(self)) {
            return IoSeq_asCString(self);
        }
        else {
            return IoSeq_asCString(IoObject_asString_(self, null));
            //IoState_error_(IoObject_state(self), null,
            //               "Unsupported type '%s'", IoObject_name(self));
            //assert(0);
        }
    }
    else static if (is(T U : U[])) {
        static if (is(Unqual!U == char)) {
            if (ISSEQ(self)) {
                static if (is(U == char)) {
                    return to!string(IoSeq_asCString(self)).dup;
                }
                else {
                    return to!string(IoSeq_asCString(self));
                }
            }
            else {
                IoState_error_(IoObject_state(self), null,
                               "Unsupported type '%s'", IoObject_name(self));
                assert(0);
            }
        }
        else {
            static assert(0, "Many array type is unimplemented. '"~T.stringof~"'");
        }
    }
    else static if (is(T == struct)) {
            writeln("name: ", to!string(IoObject_name(self)));
        if (IoDStruct!T.is_(self)) {
            writeln("pass");
            return *IoDStruct!T.data(self);
        }
        else {
            writeln("error");
            IoState_error_(IoObject_state(self), null,
                           "IoObjectTo() got '%s'. expected '%s'.",
                           IoObject_name(self), T.stringof.toStringz());
            assert(0);
        }
    }
    else static if (is(T == class)) {
        writeln("name: ", to!string(IoObject_name(self)));
        if (IoDClass!T.is_(self)) {
            writeln("pass");
            return IoDClass!T.data(self);
        }
        else {
            writeln("error");
            IoState_error_(IoObject_state(self), null,
                           "IoObjectTo() got '%s'. expected '%s'.",
                           IoObject_name(self), T.stringof.toStringz());
            assert(0);
        }
    }
    else static if (is(T == IoObject*)) {
        return self;
    }
    else {
        static assert(0, "Many type is unimplemented. '"~T.stringof~"'");
    }
}

/**
   Dの型をIoObjectに変換する
   
   Example:
       -----
       int i = 50;
       IoObject* obj = IoObjectFrom(self, i);
       IoObject* piobj = IoState_doCString(state, "1 atan * 4");
       -----
   Example:
       -----
       struct S {
           int i, j, k;
           string s;
       }
       S s;
       auto obj = IoObjectFrom(self, s);
       -----
 */
IoObject* IoObjectFrom(T)(IoState* self, T source) {
    static if (isNumeric!T) {
        return IoNumber_newWithDouble_(self, cast(double)source);
    }
    else static if (is(T == bool)) {
        return source ? self.ioTrue : self.ioFalse;
    }
    else static if (is(T U : U[])) {
        auto seq = IoSeq_newWithData_length_(self, cast(ubyte*)source.ptr, source.length);
        auto ua = cast(UArray*)(IoObject_dataPointer(seq));
        UArray_setItemType_(ua, CTypeOf!U);
        UArray_setEncoding_(ua, CEncodingOf!U);
        //IoObject_isDirty_(seq, 1);
        return seq;
    }
    else static if (is(T U : U*)) {
        static if (is(Unqual!U == IoObject)) {
            return source;
        }
        else static if (is(Unqual!U == char)) {
            return IoSeq_newWithCString_(self, source);
        }
        else static if (is(Unqual!U == void)) {
            auto seq = IoSeq_newWithData_length_(self, cast(ubyte*)&source, T.sizeof);
            auto ua = cast(UArray*)(IoObject_dataPointer(seq));
            static if (T.sizeof == 4)
                UArray_setItemType_(ua, CTYPE_uint32_t);
            else static if (T.sizeof == 8)
                UArray_setItemType_(ua, CTYPE_uint64_t);
            else
                static assert(0, "NAZO NO pointer");
            UArray_setEncoding_(ua, CENCODING_NUMBER);
            IoObject_isDirty_(seq, 1);
            return seq;
        }
        else {
            static assert(0, "'"~T.stringof~"' is not implemented");
        }
    }
    else static if (isTuple!(T)) {
        auto l = IoList_new(self);
        foreach (i, v; source.tupleof) {
            IoList_rawAppend_(l, IoObjectFrom(self, v));
        }
        return l;
    }
    else static if (is(T == struct)) {
        return IoDStruct!T.newWithData(self, source);
    }
    else static if (is(T == class)) {
        return IoDClass!T.newWithData(self, source);
    }
    else {
        return IoObjectFrom(self, "Object conversion error:"~T.stringof);
    }
}





/**
   D の struct のラッパー
 */
template IoDStruct(T) {
    /// 内部で使われます
    alias IoObject IoDObject;
    enum isStruct = is(T == struct);
    enum isClass = is(T == class);
    static if (is(T  == struct)) {
        alias T* RefType;
    }
    else static if (is(T == class)) {
        alias T  RefType;
    }
    else {
        static assert(0, "IoDStruct(IoDClass) expected class or struct.");
    }
    
    /**
       指定structのIoObjectを作成する。

       Example:
           -----
           struct S { int i, j, k; }
           auto obj = IoDStruct!S.new_(state);
           -----
    */
    IoDObject* new_(IoState* state) {
        registerProto(state);
        IoObject* proto_ = IoState_protoWithInitFunction_(state, &proto);
        return IOCLONE(proto_);
    }
    
    /// new_ の初期化子付き版。
    IoDObject* newWithData(IoState* state, T t) {
        registerProto(state);
        IoObject* proto_ = IoState_protoWithInitFunction_(state, &proto);
        auto self = IOCLONE(proto_);
        static if (isStruct) {
            *(cast(T*)IoObject_dataPointer(self)) = t;
        }
        static if (isClass) {
            IoObject_setDataPointer_(self, cast(void*)t);
        }
        GC.addRoot(IoObject_dataPointer(self));
        return self;
    }
    /// Tを取り出す。一応参照を返す。
    RefType data(IoDObject* self) {
        return cast(RefType)IoObject_dataPointer(self);
    }
    /// proto オブジェクトを取得する
    IoDObject* getProto(IoState* state) {
        return IoState_protoWithInitFunction_(state, &proto);
    }
    /**
       型チェック関数。

       Example:
           -----
           assert(IoDStruct!T.is_(IoDStruct!T.new(state))); // ok!
           assert(IoDStruct!T.is_(IoSeq_new(state))); // error!
           -----
    */
    bool is_(IoDObject* self) {
        return (IoObject_hasCloneFunc_(self, &rawClone) != 0);
    }

    // registerProto用内部変数
    private bool isRegisteredProto = false;
    /**
       プロトタイプの登録関数。

       二度以上呼んでもなにもしないが、例外も出ない。
       IoDStruct!T.new_ 等を呼ぶより前にプロトタイプが要る場合に必要。
       主に実行時に "missing proto" エラーメッセージが出現した時がそれ。
    */
    void registerProto(IoState* self) {
        if (!isRegisteredProto) {
            auto lobby = IoState_lobby(self);
            IoObject_setSlot_to_(lobby, IOSYMBOL(self, T.stringof), proto(self));
            isRegisteredProto = true;
        }
    }
    
    /// Io に登録するための内部関数群。
    extern(C) IoTag* tag(IoState* state) {
        auto tag = IoTag_newWithName_(T.stringof.toStringz());
        tag.state       = state;
        tag.cloneFunc   = &rawClone;
        tag.freeFunc    = &free;
        tag.markFunc    = &mark;
        tag.compareFunc = &compare;
        return tag;
    }
    /// ditto
    IoDObject* proto(IoState* state) {
        auto self = IoObject_new(state);
        IoObject_tag_(self, tag(state));
        //try {
            static if (__traits(compiles, T.ioAlloc(self)))
                IoObject_setDataPointer_(self, cast(void*)T.ioAlloc(self));
            else {
                static if (__traits(compiles, new T(self)))
                    IoObject_setDataPointer_(self, cast(void*)new T(self));
                else
                    IoObject_setDataPointer_(self, cast(void*)new T);
            }
            GC.addRoot(IoObject_dataPointer(self));
        /+
        }
        catch (Throwable o) {
            IoState_error_(IoObject_state(self), null,
                           "Caught D-Exception '%s'", o.toString().toStringz());
            assert(0);
        }
        +/
        IoState_registerProtoWithFunc_(state, self, &proto);

        //IoObject_addMethod_(self, IOSYMBOL(state, "clone"), &clone);
        static if (isStruct)
            IoObject_addMethod_(self, IOSYMBOL(state, "show"), &show);
        
        string makeSetterMethodName(string s) {
            return "set" ~ toupper(s[0..1]) ~ s[1..$];
        }
        
        foreach (i, field; __traits(derivedMembers, T)) {
            //pragma(msg, "field: " ~ __traits(derivedMembers, T)[i]);
            // replace "field" to "__traits(derivedMembers, T)[i]"
            static if (__traits(derivedMembers, T)[i] != "this" &&
                       __traits(derivedMembers, T)[i] != "ioClone" &&
                       __traits(derivedMembers, T)[i] != "__ctor") {
                enum isMemberFunction =
                    mixin("is(typeof(&T.init." ~ __traits(derivedMembers, T)[i] ~ ") == delegate)");
                static if (isMemberFunction) {
                    enum isStaticFunction =
                        mixin("__traits(isStaticFunction, T." ~ __traits(derivedMembers, T)[i] ~ ")");
                }
                else {
                    enum isStaticFunction = false;
                }
                enum isStatic =
                    mixin("__traits(compiles, &T." ~ __traits(derivedMembers, T)[i] ~ ")");
                enum isAccessible =
                    mixin("__traits(compiles, &T.init." ~ __traits(derivedMembers, T)[i] ~ ")");
                //pragma(msg, "isMemberFunction: "~isMemberFunction.stringof);
                //pragma(msg, "isStatic: "~isStatic.stringof);
                //pragma(msg, "isStaticFunction: "~isStaticFunction.stringof);
                //pragma(msg, "isAccessible: "~isAccessible.stringof);
                static if (isAccessible) {
                    static if (isMemberFunction && !isStaticFunction) {
                        //pragma(msg, "set method");
                        alias methodCallerFunc!(__traits(derivedMembers, T)[i]) methodCaller;
                        IoObject_addMethod_(self, IoState_symbolWithCString_(state, __traits(derivedMembers, T)[i]),
                                            &methodCaller);
                    }
                    else static if (!isStatic) {
                            //pragma(msg, "set member variable");
                        alias getterFunc!(__traits(allMembers, T)[i]) getter;
                        alias setterFunc!(__traits(allMembers, T)[i]) setter;
                        IoObject_addMethod_(self, IOSYMBOL(state, __traits(derivedMembers, T)[i]), &getter);
                        IoObject_addMethod_(self, IOSYMBOL(state, makeSetterMethodName(__traits(derivedMembers, T)[i])),
                                            &setter);
                    }
                }
            }
        }
        
        return self;
    }
    /// ditto
    extern(C) IoDObject* show(IoDObject* self) {
        auto p = cast(RefType)IoObject_dataPointer(self);
        writefln(" %s[ptr:0x%08x]", T.stringof, p);
        foreach (i, field; __traits(allMembers, T)) {
            static if (__traits(allMembers, T)[i] != "this") {
                enum isMemberFunction =
                    mixin("is(typeof(&T.init." ~ __traits(allMembers, T)[i] ~ ") == delegate)");
                enum isAccessible =
                    mixin("__traits(compiles, &T.init." ~ __traits(allMembers, T)[i] ~ ")");
                static if (!isMemberFunction && isAccessible) {
                    mixin("alias typeof(T.init."~__traits(allMembers, T)[i]~") FieldType;");
                    mixin("auto fieldvalue = p."~__traits(allMembers, T)[i]~";");
                    writefln("  %s %s \t= %s", FieldType.stringof, field, fieldvalue);
                }
            }
        }
        return self;
    }
    /// ditto
    extern(C) IoDObject* rawClone(IoDObject* self) {
        //writeln("rawClone");
        auto child = IoObject_rawClonePrimitive(self);
        try {
            static if (isStruct) {
                static if (__traits(compiles, T.init.ioClone(self))) {
                    IoObject_setDataPointer_(child, (cast(RefType)IoObject_dataPointer(self)).ioClone(self));
                }
                else {
                    IoObject_setDataPointer_(child, new T);
                    *(cast(RefType)IoObject_dataPointer(child)) = *(cast(RefType)IoObject_dataPointer(self));
                }
            }
            static if (isClass) {
                static if (__traits(compiles, T.init.ioClone(self)))
                    IoObject_setDataPointer_(child, cast(void*)(cast(RefType)IoObject_dataPointer(self)).ioClone(child));
                else
                    IoObject_setDataPointer_(child, IoObject_dataPointer(self));
            }
            GC.addRoot(IoObject_dataPointer(child));
        }
        catch (Throwable o) {
            IoState_error_(IoObject_state(self), null,
                           "Caught D-Exception '%s'", o.toString().toStringz());
            assert(0);
        }
        IoObject_tag_(child, IoObject_tag(self));
        return child;
    }
    /// ditto
    extern(C) void free(IoDObject* self) {
        GC.removeRoot(IoObject_dataPointer(self));
    }
    /// ditto
    extern(C) void mark(IoDObject* self) {
    }
    /// ditto
    extern(C) int compare(IoDObject* self) {
        return 0;
    }
    /+
    /// ditto
    extern(C) IoDObject* clone(IoDObject* self, IoObject* locals, IoMessage* m) {
        writeln("clone");
        auto child = IoObject_rawClonePrimitive(self);
        IoObject_setDataPointer_(child, new T);
        *(cast(T*)IoObject_dataPointer(child)) = *(cast(T*)IoObject_dataPointer(self));
        GC.addRoot(IoObject_dataPointer(child));
        IoObject_tag_(child, IoObject_tag(self));
        return child;
    }
    +/
    /// ditto
    extern(C) IoObject* getterFunc(string field)
        (IoObject* self, IoObject* locals, IoMessage* m) {
        auto p = IoObject_dataPointer(self);
        mixin("auto retval = (cast(RefType)p)."~field~";");
        return IoObjectFrom(IoObject_state(self), retval);
    }
    /// ditto
    extern(C) IoObject* setterFunc(string field)
        (IoObject* self, IoObject* locals, IoMessage* m) {
        auto p = IoObject_dataPointer(self);
        auto arg = IoMessage_locals_valueArgAt_(m, locals, 0);
        mixin("alias typeof((cast(RefType)p)."~field~") FT;");
        mixin("(cast(RefType)p)."~field~" = IoObjectTo!FT(arg);");
        return self;
    }
    /// ditto
    extern(C) IoObject* methodCallerFunc(string field)
        (IoObject* self, IoObject* locals, IoMessage* m) {
        try {
            auto p = IoObject_dataPointer(self);
            mixin("alias typeof(&T.init."~field~") FT;");
            alias ParameterTypeTuple!FT PT;
            static if (staticLength!PT == 0) {
                alias TypeTuple!() args;
            }
            else {
                auto args_ = ParseArguments!(Unquals!PT)(self, locals, m);
                Unquals!PT args;
                Tupleconv(args_, args);
            }
            auto casted = cast(RefType)p;
            alias ReturnType!FT RT;
            static if (is(RT == void)) {
                mixin("casted."~field~"(args);");
                return self;
            }
            else {
                mixin("auto retval = casted."~field~"(args);");
                return IoObjectFrom(IoObject_state(self), retval);
            }
        }
        catch (Throwable o) {
            IoState_error_(IoObject_state(self), m,
                           "Caught D-Exception '%s'", o.toString().toStringz());
            assert(0);
        }
    }
}
alias IoDStruct IoDClass;

/// Unqualのタプル全体適用版
template Unquals() {
    alias TypeTuple!() Unquals;
}
/// ditto
template Unquals(T1, TR...) {
    static if (TR.length == 0)
        alias TypeTuple!(Unqual!T1) Unquals;
    else
        alias TypeTuple!(Unqual!T1, Unquals!(TR)) Unquals;
}

/// std.typecons.Tuple をD内部のタプルに置き換える
void Tupleconv(T...)(ref Tuple!(T) src, ref T dst) {
    static if (staticLength!T == 0) {
    }
    else static if (staticLength!T == 1) {
        dst[0] = src[0];
    }
    else {
        foreach (i, v; T)
           dst[i] = src[i];
    }
}

/// D関数をIo関数に変換する
extern(C) IoObject* IoFunctionWrap(alias F)
(IoObject* self, IoObject* locals, IoMessage* m) {
    try {
        alias ParameterTypeTuple!F PT;
        auto args_ = ParseArguments!(Unquals!PT)(self, locals, m);
        Unquals!PT args;
        Tupleconv(args_, args);
    
        alias ReturnType!F RT;
        static if (is(RT == void)) {
            F(args);
            return self;
        }
        else {
            auto retval = F(args);
            return IoObjectFrom(IoObject_state(self), retval);
        }
    }
    catch (Exception e) {
        IoState_error_(IoObject_state(self), m,
                       "Caught D-Exception '%s'", e.toString().toStringz());
        assert(0);
    }
}

/// IoFunctionWrap内部で使われる。
/// Ioから渡された引数を解釈する。
Tuple!T ParseArguments(T...)(IoObject* self, IoObject* locals, IoMessage* m) {
    static if (staticLength!T == 0) {
        return tuple();
        //static assert(0, "argument is void");
    }
    else {
        T args;
        auto count = IoMessage_argCount(m);
        IOASSERT(self, m, count == T.length, "args error");
        auto state = IoObject_state(self);
        
        static if (staticLength!T == 1) {
            auto arg = IoMessage_locals_valueArgAt_(m, locals, 0);
            args[0] = IoObjectTo!(T[0])(arg);
        }
        else {
            foreach (i, v; T) {
                auto arg = IoMessage_locals_valueArgAt_(m, locals, i);
                args[i] = IoObjectTo!(typeof(args[i]))(arg);
            }
        }
        return tuple(args);
    }
}




